/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.mod.mixin.core.common.event.tracking.context;

import net.minecraft.util.math.BlockPos;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.WorldServer;
import net.minecraft.world.WorldType;
import net.minecraft.world.biome.Biome;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.common.event.tracking.context.SpongeProxyBlockAccess;

@Mixin(value = SpongeProxyBlockAccess.class, remap = false)
public abstract class SpongeProxyBlockAccessMixin_Forge implements IBlockAccess {

    @Shadow private WorldServer processingWorld;

    // These methods are @SideOnly(Side.CLIENT)
    // We only mix them in in SpongeForge, so that mod blocks
    // don't break when trying to call these methods on the SpongeProxyBlockAccess
    // we pass in

    @Override
    public int getCombinedLight(final BlockPos pos, final int lightValue) {
        return this.processingWorld.getCombinedLight(pos, lightValue);
    }

    @Override
    public Biome getBiome(final BlockPos pos) {
        return this.processingWorld.getBiome(pos);
    }

    @Override
    public WorldType getWorldType() {
        return this.processingWorld.getWorldType();
    }
}
